﻿//
// MainPage.xaml.cpp
// Implementation of the MainPage class.
//

#include "pch.h"
#include "MainPage.xaml.h"

#define NUM_OSCILLATORS 10

using namespace AnalogSynth;

using namespace Concurrency;
using namespace Platform;
using namespace Platform::Collections;
using namespace Windows::ApplicationModel;
using namespace Windows::Foundation;
using namespace Windows::Foundation::Collections;
using namespace Windows::Storage;
using namespace Windows::UI;
using namespace Windows::UI::Popups;
using namespace Windows::UI::Xaml;
using namespace Windows::UI::Xaml::Controls;
using namespace Windows::UI::Xaml::Controls::Primitives;
using namespace Windows::UI::Xaml::Data;
using namespace Windows::UI::Xaml::Input;
using namespace Windows::UI::Xaml::Media;
using namespace Windows::UI::Xaml::Navigation;

MainPage::MainPage()
{
	InitializeComponent();

    // Disable load button if no patches are saved
    ApplicationData^ appData = ApplicationData::Current; 
    ApplicationDataContainer^ appContainer = appData->LocalSettings;
    IMapView<String^, ApplicationDataContainer^>^ containers = appContainer->Containers;
    loadButton->IsEnabled = containers->Size > 0;

    // Create an IXAudio2 object
    HRESULT hresult = XAudio2Create(&pXAudio2);

    if (FAILED(hresult))
        ref new COMException(hresult, "XAudio2Create failure");

    // Set for debugging
    XAUDIO2_DEBUG_CONFIGURATION debugConfig = {0};
    debugConfig.TraceMask = XAUDIO2_LOG_DETAIL | XAUDIO2_LOG_WARNINGS;
    pXAudio2->SetDebugConfiguration(&debugConfig);

    // Create a mastering voice
    hresult = pXAudio2->CreateMasteringVoice(&pMasteringVoice);

    if (FAILED(hresult))
        ref new COMException(hresult, "CreateMasteringVoice failure");

    // Create volume meter effect
    IUnknown * pVolumeMeterAPO;
    XAudio2CreateVolumeMeter(&pVolumeMeterAPO);

    // Reference the effect with two structures
    XAUDIO2_EFFECT_DESCRIPTOR effectDescriptor;
    effectDescriptor.pEffect = pVolumeMeterAPO;
    effectDescriptor.InitialState = true;
    effectDescriptor.OutputChannels = 2;

    XAUDIO2_EFFECT_CHAIN effectChain;
    effectChain.EffectCount = 1;
    effectChain.pEffectDescriptors = &effectDescriptor;

    // Set the effect on the mastering voice
    pMasteringVoice->SetEffectChain(&effectChain);

    // Release the local reference to the effect
    pVolumeMeterAPO->Release();

    // Set a DispatcherTimer for polling the volume meter 
    DispatcherTimer^ timer = ref new DispatcherTimer();
    TimeSpan timespan;
    timespan.Duration = 1000000;       // 1/10th second
    timer->Interval = timespan;
    timer->Tick += ref new EventHandler<Object^>([this] (Object^, Object^)
    {
        // Poll the volume meter
        float peakLevels[2];
        float rmsLevels[2];
        XAUDIO2FX_VOLUMEMETER_LEVELS volumeMeter;
        volumeMeter.pPeakLevels = peakLevels;
        volumeMeter.pRMSLevels = rmsLevels;
        volumeMeter.ChannelCount = 2;
        pMasteringVoice->GetEffectParameters(0, &volumeMeter, sizeof(volumeMeter));

        // Set the level light to red or green
        static Brush^ redBrush = ref new SolidColorBrush(Colors::Red);
        static Brush^ greenBrush = ref new SolidColorBrush(Colors::Green);

        levelLight->Fill = (peakLevels[0] > 1 || peakLevels[1] > 1) ? redBrush : greenBrush;
    });
    timer->Start();

    // Create a sufficient number of oscillators
    for (int i = 0; i < NUM_OSCILLATORS; i++)
    {
        SynthVoice * pSynthVoice = new SynthVoice(pXAudio2.Get(), pMasteringVoice);
        allSynthVoices.push_back(pSynthVoice);
        availableSynthVoices.push_back(pSynthVoice);
    }

    Loaded += ref new RoutedEventHandler([this](Object^, RoutedEventArgs^)
    {
        // Create a singleton Patch object
        patch = Patch::Create();

        // Set event handlers for dynamic property subsets
        patch->DynamicPropertyChanged += ref new EventHandler<int>(this, &MainPage::OnDynamicPropertyChanged);
        patch->FilterEffectPropertyChanged += ref new EventHandler<int>(this, &MainPage::OnFilterEffectPropertyChanged);

        // Get the saved patch & set to DataContext
        patch->Load(nullptr);
        this->DataContext = patch;

        // Get the saved master volume
        ApplicationData^ appData = ApplicationData::Current; 
        ApplicationDataContainer^ appContainer = appData->LocalSettings;

        if (appContainer->Values->HasKey("MasterVolume"))
            volumeSlider->Value = (double)appContainer->Values->Lookup("MasterVolume");
        else
            volumeSlider->Value = 25;

        // Center the keyboard scroller
        keyboardScroller->ScrollToHorizontalOffset(max(0, (keyboardPanel->ActualWidth - this->ActualWidth) / 2));
    });

    Application::Current->Suspending += ref new SuspendingEventHandler([this](Object^, SuspendingEventArgs^)
    {
        // save the current patch
        patch->Save(nullptr);

        // Save the master volume
        ApplicationData^ appData = ApplicationData::Current; 
        ApplicationDataContainer^ appContainer = appData->LocalSettings;
        appContainer->Values->Insert("MasterVolume", volumeSlider->Value);
    });
}

void MainPage::OnDynamicPropertyChanged(Object^ sender, int args)
{
    DynamicSynthVoiceParameters params;

    params.oscillator1Waveform = (WaveformType)patch->Oscillator1Waveform;
    params.oscillator2Waveform = (WaveformType)patch->Oscillator2Waveform;

    params.oscillator1Transpose = patch->Oscillator1Transpose;
    params.oscillator2Transpose = patch->Oscillator2Transpose;
    params.oscillator2FineTune = float(pow(2.0, patch->Oscillator2FineTune / 12));

    params.oscillator1Amplitude = float(patch->Oscillator1Volume / 100);
    params.oscillator2Amplitude = float(patch->Oscillator2Volume / 100);

    for (SynthVoice * pSynthVoice : allSynthVoices)
        pSynthVoice->SetParameters(&params);
}

void MainPage::OnFilterEffectPropertyChanged(Object^ sender, int args)
{
    DynamicFilterParameters params;
    params.filterOn = patch->FilterOn;
    params.filterCutoffFrequency = float(patch->FilterCutoffFrequency);
    params.filterEmphasis = float(patch->FilterEmphasis);

    for (SynthVoice * pSynthVoice : allSynthVoices)
        pSynthVoice->SetFilterParameters(&params);
}

void MainPage::OnKeyIsPressedChanged(Object^ sender, KeyPressedArgs^ args)
{
    Key^ key = dynamic_cast<Key^>(sender);
    int keyNum = args->KeyNumber;

    if (args->IsPressed)
    {
        if (availableSynthVoices.size() > 0)
        {
            // Get SynthVoice from end of availableSynthVoices
            SynthVoice * pSynthVoice = availableSynthVoices.back();
            availableSynthVoices.pop_back();

            // Set up a TriggeredSynthVoiceParameters structure
            TriggeredSynthVoiceParameters params;
            params.triggerFrequency = float(440 * pow(2, (keyNum % 1000 - 69) / 12.0));

            // Amplitude envelope
            float attackTime = float(pow(10, patch->LoudnessAttack));
            float decayTime = float(pow(10, patch->LoudnessDecay));
            float sustainLevel = float(patch->LoudnessSustain / 100);
            float releaseTime = float(pow(10, patch->LoudnessRelease));

            params.amplitudeEnvelope.baseLevel = 0;
            params.amplitudeEnvelope.attackRate = float(1 / attackTime);
            params.amplitudeEnvelope.peakLevel = 1;
            params.amplitudeEnvelope.decayRate = float((sustainLevel - 1) / decayTime);
            params.amplitudeEnvelope.sustainLevel = sustainLevel;
            params.amplitudeEnvelope.releaseTime = releaseTime;

            // Filter parameters
            params.filterParameters.filterOn = patch->FilterOn;
            params.filterParameters.filterCutoffFrequency = float(patch->FilterCutoffFrequency);
            params.filterParameters.filterEmphasis = float(patch->FilterEmphasis);

            // Filter envelope
            float baseLevel = 1;
            float peakLevel = baseLevel + float(patch->FilterEnvelope / 6.67);        // ie, 4 octaves
            sustainLevel = baseLevel + float(patch->FilterEnvelope * patch->FilterSustain / 667);

            attackTime = float(pow(10, patch->FilterAttack));
            decayTime = float(pow(10, patch->FilterDecay));
            releaseTime = float(pow(10, patch->FilterRelease));

            params.filterEnvelope.baseLevel = baseLevel;
            params.filterEnvelope.attackRate = float((peakLevel - baseLevel) / attackTime);
            params.filterEnvelope.peakLevel = peakLevel;
            params.filterEnvelope.decayRate = float((sustainLevel - peakLevel) / decayTime);
            params.filterEnvelope.sustainLevel = sustainLevel;
            params.filterEnvelope.releaseTime = releaseTime;

            // Trigger the voice
            pSynthVoice->Trigger(&params);

            // Add the voice to the "playing" collection
            playingSynthVoices[keyNum] = pSynthVoice;
        }
    }
    else
    {
        // Get the playing SynthVoice
        SynthVoice * pSynthVoice = playingSynthVoices[keyNum];

        if (pSynthVoice != nullptr)
        {
            // Signal a release
            pSynthVoice->Release();

            // At this point, the key has been released but the voice is not
            //  yet completed because the release segment of the envelope
            //  is still in progress.

            // For that reason, put the voice at the beginning of the 
            //  collection rather than the end to avoid it being reused
            //  immediately.
            availableSynthVoices.insert(availableSynthVoices.begin(), pSynthVoice);

            // Remove it from playingSynthVoices collection
            playingSynthVoices.erase(keyNum);
        }
    }
}

// Shift keyboard in accordance with ScrollViewer
void MainPage::OnScrollViewerViewChanged(Object^ sender, ScrollViewerViewChangedEventArgs^ args)
{
    ScrollViewer^ scrollViewer = safe_cast<ScrollViewer^>(sender);
    keyboardTranslateTransform->X = -scrollViewer->HorizontalOffset;
}

void MainPage::OnSaveButtonClick(Object^ sender, RoutedEventArgs^ args)
{
    // Come up with unique name
    ApplicationData^ appData = ApplicationData::Current; 
    ApplicationDataContainer^ appContainer = appData->LocalSettings;
    IMapView<String^, ApplicationDataContainer^>^ containers = appContainer->Containers;
    int patchNumber = 0;
    bool isUnique;

    do
    {
        patchNumber += 1;
        isUnique = true;
        String^ patchName = "patch " + patchNumber;

        for each (IKeyValuePair<String^, ApplicationDataContainer^>^ kvp in containers)
        {
            if (kvp->Key == patchName)
            {
                isUnique = false;
                break;
            }
        }
    }
    while (!isUnique);

    // Create Save dialog
    PatchSaveDialog^ dlg = ref new PatchSaveDialog("patch " + patchNumber.ToString());

    // Create Popup
    Popup^ popup = ref new Popup();
    popup->Child = dlg;
    popup->IsLightDismissEnabled = true;
    popup->IsOpen = true;

    dlg->SizeChanged += ref new SizeChangedEventHandler(
        [this, popup, dlg] (Object^, SizeChangedEventArgs^ args)
    {
        popup->HorizontalOffset = this->ActualWidth - dlg->ActualWidth - 50;
        popup->VerticalOffset = 50;
    });

    dlg->Save += ref new EventHandler<String^>(
        [this, popup, dlg, appContainer] (Object^, String^ args)
    {
        popup->IsOpen = false;

        if (appContainer->Containers->HasKey(args))
        {
            MessageDialog^ msgdlg = ref new MessageDialog("Replace existing patch named '" + args + "'?");
            msgdlg->Commands->Append(ref new UICommand("No", nullptr, false));
            msgdlg->Commands->Append(ref new UICommand("Yes", nullptr, true));
            msgdlg->CancelCommandIndex = 0;
            msgdlg->DefaultCommandIndex = 1;

            task<IUICommand^> messageTask = create_task(msgdlg->ShowAsync());
            messageTask.then([this, args] (IUICommand^ command)
            {
                if ((bool)command->Id)
                {
                    patch->Save(args);
                    loadButton->IsEnabled = true;   // because at least one item
                };
            });
        }
        else
        {
            patch->Save(args);
            loadButton->IsEnabled = true;   // because at least one item
        }
    });
}

void MainPage::OnLoadButtonClick(Object^ sender, RoutedEventArgs^ args)
{
    // Enumerate containers for dialog box ListBox
    ApplicationData^ appData = ApplicationData::Current; 
    ApplicationDataContainer^ appContainer = appData->LocalSettings;
    IMapView<String^, ApplicationDataContainer^>^ containers = appContainer->Containers;

    Vector<String^>^ names = ref new Vector<String^>();

    for each (IKeyValuePair<String^, ApplicationDataContainer^>^ kvp in containers)
        names->Append(kvp->Key);

    // Create Load dialog
    PatchLoadDialog^ dlg = ref new PatchLoadDialog(names->GetView());

    // Create Popup
    Popup^ popup = ref new Popup();
    popup->Child = dlg;
    popup->IsLightDismissEnabled = true;
    popup->IsOpen = true;

    dlg->SizeChanged += ref new SizeChangedEventHandler(
        [this, popup, dlg] (Object^, SizeChangedEventArgs^ args)
    {
        popup->HorizontalOffset = this->ActualWidth - dlg->ActualWidth - 50;
        popup->VerticalOffset = 50;
    });

    dlg->Load += ref new EventHandler<String^>(
        [this, popup, dlg] (Object^, String^ args)
    {
        popup->IsOpen = false;
        patch->Load(args);
    });
}

void MainPage::OnVolumeSliderValueChanged(Object^ sender, RangeBaseValueChangedEventArgs^ args)
{
    HRESULT hresult = pMasteringVoice->SetVolume(float(args->NewValue) / 100);

    if (FAILED(hresult))
        ref new COMException(hresult, "pMasteringVoice->SetVolume failure");
}
